@@ -31,6 +31,7 @@
 import io
 import sys
 import ctypes
+import platform
 
 from pycriu import images
 from . import elf
@@ -54,6 +55,7 @@ status = {
     "VMA_AREA_VVAR": 1 << 12,
     "VMA_AREA_AIORING": 1 << 13,
     "VMA_AREA_MEMFD": 1 << 14,
+    "VMA_AREA_UPROBES": 1 << 17,
     "VMA_AREA_UNSUPP": 1 << 31
 }
 
@@ -94,8 +96,13 @@ class coredump:
             buf.write(b"\0" * (8 - len(note.owner)))
             buf.write(note.data)
 
-        offset = ctypes.sizeof(elf.Elf64_Ehdr())
-        offset += (len(self.vmas) + 1) * ctypes.sizeof(elf.Elf64_Phdr())
+        bits = platform.architecture()[0]  # 32 or 64 bits
+
+        ehdr = {"32bit": elf.Elf32_Ehdr, "64bit": elf.Elf64_Ehdr}
+        phdr = {"32bit": elf.Elf32_Phdr, "64bit": elf.Elf64_Phdr}
+
+        offset = ctypes.sizeof(ehdr[bits]())
+        offset += (len(self.vmas) + 1) * ctypes.sizeof(phdr[bits]())
 
         filesz = 0
         for note in self.notes:
@@ -130,6 +137,20 @@ class coredump_generator:
     reg_files = None  # reg-files;
     pagemaps = {}  # pagemap by pid;
 
+    # thread info key based on the current arch
+    thread_info_key = {
+        "aarch64": "ti_aarch64",
+        "armv7l": "ti_arm",
+        "x86_64": "thread_info",
+    }
+
+    machine = platform.machine()  # current arch
+    bits = platform.architecture()[0]  # 32 or 64 bits
+
+    ehdr = {"32bit": elf.Elf32_Ehdr, "64bit": elf.Elf64_Ehdr}  # 32 or 64 bits Ehdr
+    nhdr = {"32bit": elf.Elf32_Nhdr, "64bit": elf.Elf64_Nhdr}  # 32 or 64 bits Nhdr
+    phdr = {"32bit": elf.Elf32_Phdr, "64bit": elf.Elf64_Phdr}  # 32 or 64 bits Phdr
+
     def _img_open_and_strip(self, name, single=False, pid=None):
         """
         Load criu image and strip it from magic and redundant list.
@@ -201,44 +222,62 @@ class coredump_generator:
         """
         Generate elf header for process pid with program headers phdrs.
         """
-        ehdr = elf.Elf64_Ehdr()
+        ei_class = {"32bit": elf.ELFCLASS32, "64bit": elf.ELFCLASS64}
+
+        ehdr = self.ehdr[self.bits]()
 
         ctypes.memset(ctypes.addressof(ehdr), 0, ctypes.sizeof(ehdr))
         ehdr.e_ident[elf.EI_MAG0] = elf.ELFMAG0
         ehdr.e_ident[elf.EI_MAG1] = elf.ELFMAG1
         ehdr.e_ident[elf.EI_MAG2] = elf.ELFMAG2
         ehdr.e_ident[elf.EI_MAG3] = elf.ELFMAG3
-        ehdr.e_ident[elf.EI_CLASS] = elf.ELFCLASS64
+        ehdr.e_ident[elf.EI_CLASS] = ei_class[self.bits]
         ehdr.e_ident[elf.EI_DATA] = elf.ELFDATA2LSB
         ehdr.e_ident[elf.EI_VERSION] = elf.EV_CURRENT
 
+        if self.machine == "armv7l":
+            ehdr.e_ident[elf.EI_OSABI] = elf.ELFOSABI_ARM
+        else:
+            ehdr.e_ident[elf.EI_OSABI] = elf.ELFOSABI_NONE
+
         ehdr.e_type = elf.ET_CORE
-        ehdr.e_machine = elf.EM_X86_64
+        ehdr.e_machine = self._get_e_machine()
         ehdr.e_version = elf.EV_CURRENT
-        ehdr.e_phoff = ctypes.sizeof(elf.Elf64_Ehdr())
-        ehdr.e_ehsize = ctypes.sizeof(elf.Elf64_Ehdr())
-        ehdr.e_phentsize = ctypes.sizeof(elf.Elf64_Phdr())
+        ehdr.e_phoff = ctypes.sizeof(self.ehdr[self.bits]())
+        ehdr.e_ehsize = ctypes.sizeof(self.ehdr[self.bits]())
+        ehdr.e_phentsize = ctypes.sizeof(self.phdr[self.bits]())
         # FIXME Case len(phdrs) > PN_XNUM should be handled properly.
         # See fs/binfmt_elf.c from linux kernel.
         ehdr.e_phnum = len(phdrs)
 
         return ehdr
 
+    def _get_e_machine(self):
+        """
+        Get the e_machine field based on the current architecture.
+        """
+        e_machine_dict = {
+            "aarch64": elf.EM_AARCH64,
+            "armv7l": elf.EM_ARM,
+            "x86_64": elf.EM_X86_64,
+        }
+        return e_machine_dict[self.machine]
+
     def _gen_phdrs(self, pid, notes, vmas):
         """
         Generate program headers for process pid.
         """
         phdrs = []
 
-        offset = ctypes.sizeof(elf.Elf64_Ehdr())
-        offset += (len(vmas) + 1) * ctypes.sizeof(elf.Elf64_Phdr())
+        offset = ctypes.sizeof(self.ehdr[self.bits]())
+        offset += (len(vmas) + 1) * ctypes.sizeof(self.phdr[self.bits]())
 
         filesz = 0
         for note in notes:
             filesz += ctypes.sizeof(note.nhdr) + ctypes.sizeof(note.data) + 8
 
         # PT_NOTE
-        phdr = elf.Elf64_Phdr()
+        phdr = self.phdr[self.bits]()
         ctypes.memset(ctypes.addressof(phdr), 0, ctypes.sizeof(phdr))
         phdr.p_type = elf.PT_NOTE
         phdr.p_offset = offset
@@ -258,7 +297,7 @@ class coredump_generator:
         for vma in vmas:
             offset += filesz
             filesz = vma.filesz
-            phdr = elf.Elf64_Phdr()
+            phdr = self.phdr[self.bits]()
             ctypes.memset(ctypes.addressof(phdr), 0, ctypes.sizeof(phdr))
             phdr.p_type = elf.PT_LOAD
             phdr.p_align = PAGESIZE
@@ -315,7 +354,7 @@ class coredump_generator:
         prpsinfo.pr_psargs = self._gen_cmdline(pid)[:80]
         prpsinfo.pr_fname = core["tc"]["comm"].encode()
 
-        nhdr = elf.Elf64_Nhdr()
+        nhdr = self.nhdr[self.bits]()
         nhdr.n_namesz = 5
         nhdr.n_descsz = ctypes.sizeof(elf.elf_prpsinfo())
         nhdr.n_type = elf.NT_PRPSINFO
@@ -332,7 +371,7 @@ class coredump_generator:
         Generate NT_PRSTATUS note for thread tid of process pid.
         """
         core = self.cores[tid]
-        regs = core["thread_info"]["gpregs"]
+        regs = self._get_gpregs(core)
         pstree = self.pstree[pid]
 
         prstatus = elf.elf_prstatus()
@@ -345,35 +384,9 @@ class coredump_generator:
         prstatus.pr_pgrp = pstree["pgid"]
         prstatus.pr_sid = pstree["sid"]
 
-        prstatus.pr_reg.r15 = regs["r15"]
-        prstatus.pr_reg.r14 = regs["r14"]
-        prstatus.pr_reg.r13 = regs["r13"]
-        prstatus.pr_reg.r12 = regs["r12"]
-        prstatus.pr_reg.rbp = regs["bp"]
-        prstatus.pr_reg.rbx = regs["bx"]
-        prstatus.pr_reg.r11 = regs["r11"]
-        prstatus.pr_reg.r10 = regs["r10"]
-        prstatus.pr_reg.r9 = regs["r9"]
-        prstatus.pr_reg.r8 = regs["r8"]
-        prstatus.pr_reg.rax = regs["ax"]
-        prstatus.pr_reg.rcx = regs["cx"]
-        prstatus.pr_reg.rdx = regs["dx"]
-        prstatus.pr_reg.rsi = regs["si"]
-        prstatus.pr_reg.rdi = regs["di"]
-        prstatus.pr_reg.orig_rax = regs["orig_ax"]
-        prstatus.pr_reg.rip = regs["ip"]
-        prstatus.pr_reg.cs = regs["cs"]
-        prstatus.pr_reg.eflags = regs["flags"]
-        prstatus.pr_reg.rsp = regs["sp"]
-        prstatus.pr_reg.ss = regs["ss"]
-        prstatus.pr_reg.fs_base = regs["fs_base"]
-        prstatus.pr_reg.gs_base = regs["gs_base"]
-        prstatus.pr_reg.ds = regs["ds"]
-        prstatus.pr_reg.es = regs["es"]
-        prstatus.pr_reg.fs = regs["fs"]
-        prstatus.pr_reg.gs = regs["gs"]
+        self._set_pr_regset(prstatus.pr_reg, regs)
 
-        nhdr = elf.Elf64_Nhdr()
+        nhdr = self.nhdr[self.bits]()
         nhdr.n_namesz = 5
         nhdr.n_descsz = ctypes.sizeof(elf.elf_prstatus())
         nhdr.n_type = elf.NT_PRSTATUS
@@ -385,28 +398,83 @@ class coredump_generator:
 
         return note
 
+    def _get_gpregs(self, core):
+        """
+        Get the general purpose registers based on the current architecture.
+        """
+        thread_info_key = self.thread_info_key[self.machine]
+        thread_info = core[thread_info_key]
+
+        return thread_info["gpregs"]
+
+    def _set_pr_regset(self, pr_reg, regs):
+        """
+        Set the pr_reg struct based on the current architecture.
+        """
+        if self.machine == "aarch64":
+            pr_reg.regs = (ctypes.c_ulonglong * len(regs["regs"]))(*regs["regs"])
+            pr_reg.sp = regs["sp"]
+            pr_reg.pc = regs["pc"]
+            pr_reg.pstate = regs["pstate"]
+        elif self.machine == "armv7l":
+            pr_reg.r0 = regs["r0"]
+            pr_reg.r1 = regs["r1"]
+            pr_reg.r2 = regs["r2"]
+            pr_reg.r3 = regs["r3"]
+            pr_reg.r4 = regs["r4"]
+            pr_reg.r5 = regs["r5"]
+            pr_reg.r6 = regs["r6"]
+            pr_reg.r7 = regs["r7"]
+            pr_reg.r8 = regs["r8"]
+            pr_reg.r9 = regs["r9"]
+            pr_reg.r10 = regs["r10"]
+            pr_reg.fp = regs["fp"]
+            pr_reg.ip = regs["ip"]
+            pr_reg.sp = regs["sp"]
+            pr_reg.lr = regs["lr"]
+            pr_reg.pc = regs["pc"]
+            pr_reg.cpsr = regs["cpsr"]
+            pr_reg.orig_r0 = regs["orig_r0"]
+        elif self.machine == "x86_64":
+            pr_reg.r15 = regs["r15"]
+            pr_reg.r14 = regs["r14"]
+            pr_reg.r13 = regs["r13"]
+            pr_reg.r12 = regs["r12"]
+            pr_reg.rbp = regs["bp"]
+            pr_reg.rbx = regs["bx"]
+            pr_reg.r11 = regs["r11"]
+            pr_reg.r10 = regs["r10"]
+            pr_reg.r9 = regs["r9"]
+            pr_reg.r8 = regs["r8"]
+            pr_reg.rax = regs["ax"]
+            pr_reg.rcx = regs["cx"]
+            pr_reg.rdx = regs["dx"]
+            pr_reg.rsi = regs["si"]
+            pr_reg.rdi = regs["di"]
+            pr_reg.orig_rax = regs["orig_ax"]
+            pr_reg.rip = regs["ip"]
+            pr_reg.cs = regs["cs"]
+            pr_reg.eflags = regs["flags"]
+            pr_reg.rsp = regs["sp"]
+            pr_reg.ss = regs["ss"]
+            pr_reg.fs_base = regs["fs_base"]
+            pr_reg.gs_base = regs["gs_base"]
+            pr_reg.ds = regs["ds"]
+            pr_reg.es = regs["es"]
+            pr_reg.fs = regs["fs"]
+            pr_reg.gs = regs["gs"]
+
     def _gen_fpregset(self, pid, tid):
         """
         Generate NT_FPREGSET note for thread tid of process pid.
         """
         core = self.cores[tid]
-        regs = core["thread_info"]["fpregs"]
+        regs = self._get_fpregs(core)
 
         fpregset = elf.elf_fpregset_t()
         ctypes.memset(ctypes.addressof(fpregset), 0, ctypes.sizeof(fpregset))
 
-        fpregset.cwd = regs["cwd"]
-        fpregset.swd = regs["swd"]
-        fpregset.ftw = regs["twd"]
-        fpregset.fop = regs["fop"]
-        fpregset.rip = regs["rip"]
-        fpregset.rdp = regs["rdp"]
-        fpregset.mxcsr = regs["mxcsr"]
-        fpregset.mxcr_mask = regs["mxcsr_mask"]
-        fpregset.st_space = (ctypes.c_uint * len(regs["st_space"]))(
-            *regs["st_space"])
-        fpregset.xmm_space = (ctypes.c_uint * len(regs["xmm_space"]))(
-            *regs["xmm_space"])
+        self._set_fpregset(fpregset, regs)
 
         nhdr = elf.Elf64_Nhdr()
         nhdr.n_namesz = 5
@@ -420,6 +488,86 @@ class coredump_generator:
 
         return note
 
+    def _get_fpregs(self, core):
+        """
+        Get the floating point register dictionary based on the current architecture.
+        """
+        fpregs_key_dict = {"aarch64": "fpsimd", "x86_64": "fpregs"}
+        fpregs_key = fpregs_key_dict[self.machine]
+
+        thread_info_key = self.thread_info_key[self.machine]
+
+        return core[thread_info_key][fpregs_key]
+
+    def _set_fpregset(self, fpregset, regs):
+        """
+        Set the fpregset struct based on the current architecture.
+        """
+        if self.machine == "aarch64":
+            fpregset.vregs = (ctypes.c_ulonglong * len(regs["vregs"]))(*regs["vregs"])
+            fpregset.fpsr = regs["fpsr"]
+            fpregset.fpcr = regs["fpcr"]
+        elif self.machine == "x86_64":
+            fpregset.cwd = regs["cwd"]
+            fpregset.swd = regs["swd"]
+            fpregset.ftw = regs["twd"]
+            fpregset.fop = regs["fop"]
+            fpregset.rip = regs["rip"]
+            fpregset.rdp = regs["rdp"]
+            fpregset.mxcsr = regs["mxcsr"]
+            fpregset.mxcr_mask = regs["mxcsr_mask"]
+            fpregset.st_space = (ctypes.c_uint * len(regs["st_space"]))(
+                *regs["st_space"])
+            fpregset.xmm_space = (ctypes.c_uint * len(regs["xmm_space"]))(
+                *regs["xmm_space"])
+
+    def _gen_arm_tls(self, tid):
+        """
+        Generate NT_ARM_TLS note for thread tid of process pid.
+        """
+        core = self.cores[tid]
+        tls = ctypes.c_ulonglong(core["ti_aarch64"]["tls"])
+
+        nhdr = elf.Elf64_Nhdr()
+        nhdr.n_namesz = 6
+        nhdr.n_descsz = ctypes.sizeof(ctypes.c_ulonglong)
+        nhdr.n_type = elf.NT_ARM_TLS
+
+        note = elf_note()
+        note.data = tls
+        note.owner = b"LINUX"
+        note.nhdr = nhdr
+
+        return note
+
+    def _gen_arm_vfp(self, tid):
+        """
+        Generate NT_ARM_VFP note for thread tid of process pid.
+        """
+        core = self.cores[tid]
+        fpstate = core["ti_arm"]["fpstate"]
+
+        data = elf.vfp_hard_struct()
+        ctypes.memset(ctypes.addressof(data), 0, ctypes.sizeof(data))
+
+        data.vfp_regs = (ctypes.c_uint64 * len(fpstate["vfp_regs"]))(*fpstate["vfp_regs"])
+        data.fpexc = fpstate["fpexc"]
+        data.fpscr = fpstate["fpscr"]
+        data.fpinst = fpstate["fpinst"]
+        data.fpinst2 = fpstate["fpinst2"]
+
+        nhdr = elf.Elf32_Nhdr()
+        nhdr.n_namesz = 6
+        nhdr.n_descsz = ctypes.sizeof(data)
+        nhdr.n_type = elf.NT_ARM_VFP
+
+        note = elf_note()
+        note.data = data
+        note.owner = b"LINUX"
+        note.nhdr = nhdr
+
+        return note
+
     def _gen_x86_xstate(self, pid, tid):
         """
         Generate NT_X86_XSTATE note for thread tid of process pid.
@@ -469,7 +617,7 @@ class coredump_generator:
         # FIXME zeroify everything for now
         ctypes.memset(ctypes.addressof(siginfo), 0, ctypes.sizeof(siginfo))
 
-        nhdr = elf.Elf64_Nhdr()
+        nhdr = self.nhdr[self.bits]()
         nhdr.n_namesz = 5
         nhdr.n_descsz = ctypes.sizeof(elf.siginfo_t())
         nhdr.n_type = elf.NT_SIGINFO
@@ -488,17 +636,22 @@ class coredump_generator:
         mm = self.mms[pid]
         num_auxv = len(mm["mm_saved_auxv"]) // 2
 
-        class elf_auxv(ctypes.Structure):
+        class elf32_auxv(ctypes.Structure):
+            _fields_ = [("auxv", elf.Elf32_auxv_t * num_auxv)]
+
+        class elf64_auxv(ctypes.Structure):
             _fields_ = [("auxv", elf.Elf64_auxv_t * num_auxv)]
 
-        auxv = elf_auxv()
+        elf_auxv = {"32bit": elf32_auxv(), "64bit": elf64_auxv()}
+
+        auxv = elf_auxv[self.bits]
         for i in range(num_auxv):
             auxv.auxv[i].a_type = mm["mm_saved_auxv"][i]
             auxv.auxv[i].a_val = mm["mm_saved_auxv"][i + 1]
 
-        nhdr = elf.Elf64_Nhdr()
+        nhdr = self.nhdr[self.bits]()
         nhdr.n_namesz = 5
-        nhdr.n_descsz = ctypes.sizeof(elf_auxv())
+        nhdr.n_descsz = ctypes.sizeof(elf_auxv[self.bits])
         nhdr.n_type = elf.NT_AUXV
 
         note = elf_note()
@@ -575,7 +728,7 @@ class coredump_generator:
             setattr(data, "file_ofs" + str(i), info.file_ofs)
             setattr(data, "name" + str(i), info.name.encode())
 
-        nhdr = elf.Elf64_Nhdr()
+        nhdr = self.nhdr[self.bits]()
 
         nhdr.n_namesz = 5  # strlen + 1
         nhdr.n_descsz = ctypes.sizeof(elf_files())
@@ -592,9 +745,15 @@ class coredump_generator:
         notes = []
 
         notes.append(self._gen_prstatus(pid, tid))
-        notes.append(self._gen_fpregset(pid, tid))
-        notes.append(self._gen_x86_xstate(pid, tid))
+        if self.machine != "armv7l":
+            notes.append(self._gen_fpregset(pid, tid))
         notes.append(self._gen_siginfo(pid, tid))
+        if self.machine == "aarch64":
+            notes.append(self._gen_arm_tls(tid))
+        elif self.machine == "armv7l":
+            notes.append(self._gen_arm_vfp(tid))
+        elif self.machine == "x86_64":
+            notes.append(self._gen_x86_xstate(pid, tid))
 
         return notes
 
@@ -635,7 +794,9 @@ class coredump_generator:
         off = 0  # in pages
         for m in pagemap[1:]:
             found = False
-            for i in range(m["nr_pages"]):
+            num_pages = m.get("nr_pages", m["compat_nr_pages"])
+
+            for i in range(num_pages):
                 if m["vaddr"] + i * PAGESIZE == page_no * PAGESIZE:
                     found = True
                     break
