@@ -54,19 +54,17 @@ void flush_early_log_to_stderr(void)
 	flush_early_log_buffer(STDERR_FILENO);
 }
 
-static int image_dir_mode(char *argv[], int optind)
+static int image_dir_mode(void)
 {
 	switch (opts.mode) {
 	case CR_DUMP:
 		/* fallthrough */
+	case CR_CPUINFO_DUMP:
+		/* fallthrough */
 	case CR_PRE_DUMP:
 		return O_DUMP;
 	case CR_RESTORE:
 		return O_RSTR;
-	case CR_CPUINFO:
-		if (!strcmp(argv[optind + 1], "dump"))
-			return O_DUMP;
-		/* fallthrough */
 	default:
 		return -1;
 	}
@@ -76,36 +74,55 @@ static int image_dir_mode(char *argv[], int optind)
 	return -1;
 }
 
-static int parse_criu_mode(char *mode)
-{
-	if (!strcmp(mode, "dump"))
-		opts.mode = CR_DUMP;
-	else if (!strcmp(mode, "pre-dump"))
-		opts.mode = CR_PRE_DUMP;
-	else if (!strcmp(mode, "restore"))
-		opts.mode = CR_RESTORE;
-	else if (!strcmp(mode, "lazy-pages"))
-		opts.mode = CR_LAZY_PAGES;
-	else if (!strcmp(mode, "check"))
-		opts.mode = CR_CHECK;
-	else if (!strcmp(mode, "page-server"))
-		opts.mode = CR_PAGE_SERVER;
-	else if (!strcmp(mode, "service"))
-		opts.mode = CR_SERVICE;
-	else if (!strcmp(mode, "swrk"))
-		opts.mode = CR_SWRK;
-	else if (!strcmp(mode, "dedup"))
-		opts.mode = CR_DEDUP;
-	else if (!strcmp(mode, "cpuinfo"))
-		opts.mode = CR_CPUINFO;
-	else if (!strcmp(mode, "exec"))
-		opts.mode = CR_EXEC_DEPRECATED;
-	else if (!strcmp(mode, "show"))
-		opts.mode = CR_SHOW_DEPRECATED;
-	else
-		return -1;
+struct {
+	char *cmd;
+	int mode;
+} commands[] = {
+	{ "dump", CR_DUMP },
+	{ "pre-dump", CR_PRE_DUMP },
+	{ "restore", CR_RESTORE },
+	{ "lazy-pages", CR_LAZY_PAGES },
+	{ "check", CR_CHECK },
+	{ "page-server", CR_PAGE_SERVER },
+	{ "service", CR_SERVICE },
+	{ "swrk", CR_SWRK },
+	{ "dedup", CR_DEDUP },
+	{ "exec", CR_EXEC_DEPRECATED },
+	{ "show", CR_SHOW_DEPRECATED },
+};
 
-	return 0;
+static int parse_criu_mode(int argc, char **argv, int *optind)
+{
+	char *cmd = argv[*optind];
+	bool has_sub_command = (argc - *optind) > 1;
+	char *subcommand = has_sub_command ? argv[*optind + 1] : NULL;
+	int i;
+
+	for (i = 0; i < ARRAY_SIZE(commands); i++) {
+		if (strcmp(cmd, commands[i].cmd))
+			continue;
+		opts.mode = commands[i].mode;
+		return 0;
+	}
+
+	if (!strcmp(cmd, "cpuinfo")) {
+		if (subcommand == NULL) {
+			pr_err("cpuinfo requires an action: dump or check\n");
+			return -1;
+		}
+		if (!strcmp(subcommand, "dump"))
+			opts.mode = CR_CPUINFO_DUMP;
+		else if (!strcmp(subcommand, "check"))
+			opts.mode = CR_CPUINFO_CHECK;
+		else {
+			pr_err("unknown cpuinfo sub-command: %s\n", subcommand);
+			return -1;
+		}
+		(*optind)++;
+		return 0;
+	}
+	pr_err("unknown command: %s\n", argv[*optind]);
+	return -1;
 }
 
 int main(int argc, char *argv[], char *envp[])
@@ -115,6 +132,7 @@ int main(int argc, char *argv[], char *envp[])
 	bool has_exec_cmd = false;
 	bool has_sub_command;
 	int state = PARSING_GLOBAL_CONF;
+	char *cmd;
 
 	BUILD_BUG_ON(CTL_32 != SYSCTL_TYPE__CTL_32);
 	BUILD_BUG_ON(__CTL_STR != SYSCTL_TYPE__CTL_STR);
@@ -165,11 +183,18 @@ int main(int argc, char *argv[], char *envp[])
 		return 1;
 	}
 
-	if (parse_criu_mode(argv[optind])) {
-		pr_err("unknown command: %s\n", argv[optind]);
+	cmd = argv[optind];
+	ret = parse_criu_mode(argc, argv, &optind);
+	if (ret)
 		goto usage;
-	}
 
+	/*
+	 * util_init initializes criu_run_id and compel_run_id so that sockets
+	 * are generated with an unique name identifying the specific process
+	 * even in cases where multiple processes with the same pid in
+	 * different pid namespaces are sharing the same network namespace.
+	 */
+	util_init();
 	if (opts.mode == CR_SWRK) {
 		if (argc != optind + 2) {
 			fprintf(stderr, "Usage: criu swrk <fd>\n");
@@ -217,25 +242,19 @@ int main(int argc, char *argv[], char *envp[])
 			return 1;
 		memcpy(opts.exec_cmd, &argv[optind + 1], (argc - optind - 1) * sizeof(char *));
 		opts.exec_cmd[argc - optind - 1] = NULL;
-	} else {
-		/* No subcommands except for cpuinfo and restore --exec-cmd */
-		if (opts.mode != CR_CPUINFO && has_sub_command) {
-			pr_err("excessive parameter%s for command %s\n", (argc - optind) > 2 ? "s" : "", argv[optind]);
-			goto usage;
-		} else if (opts.mode == CR_CPUINFO && !has_sub_command) {
-			pr_err("cpuinfo requires an action: dump or check\n");
-			goto usage;
-		}
+	} else if (has_sub_command) {
+		pr_err("excessive parameter%s for command %s\n", (argc - optind) > 2 ? "s" : "", cmd);
+		goto usage;
 	}
 
-	if (opts.stream && image_dir_mode(argv, optind) == -1) {
-		pr_err("--stream cannot be used with the %s command\n", argv[optind]);
+	if (opts.stream && image_dir_mode() == -1) {
+		pr_err("--stream cannot be used with the %s command\n", cmd);
 		goto usage;
 	}
 
 	/* We must not open imgs dir, if service is called */
 	if (opts.mode != CR_SERVICE) {
-		ret = open_image_dir(opts.imgs_dir, image_dir_mode(argv, optind));
+		ret = open_image_dir(opts.imgs_dir, image_dir_mode());
 		if (ret < 0) {
 			pr_err("Couldn't open image dir %s\n", opts.imgs_dir);
 			return 1;
@@ -254,8 +273,6 @@ int main(int argc, char *argv[], char *envp[])
 		return 1;
 	}
 
-	util_init();
-
 	if (log_init(opts.output))
 		return 1;
 
@@ -292,14 +309,13 @@ int main(int argc, char *argv[], char *envp[])
 	if (opts.img_parent)
 		pr_info("Will do snapshot from %s\n", opts.img_parent);
 
-	if (opts.mode == CR_DUMP) {
+	switch (opts.mode) {
+	case CR_DUMP:
 		if (!opts.tree_id)
 			goto opt_pid_missing;
 
 		return cr_dump_tasks(opts.tree_id);
-	}
-
-	if (opts.mode == CR_PRE_DUMP) {
+	case CR_PRE_DUMP:
 		if (!opts.tree_id)
 			goto opt_pid_missing;
 
@@ -309,9 +325,7 @@ int main(int argc, char *argv[], char *envp[])
 		}
 
 		return cr_pre_dump_tasks(opts.tree_id) != 0;
-	}
-
-	if (opts.mode == CR_RESTORE) {
+	case CR_RESTORE:
 		if (opts.tree_id)
 			pr_warn("Using -t with criu restore is obsoleted\n");
 
@@ -324,46 +338,41 @@ int main(int argc, char *argv[], char *envp[])
 		}
 
 		return ret != 0;
-	}
 
-	if (opts.mode == CR_LAZY_PAGES)
+	case CR_LAZY_PAGES:
 		return cr_lazy_pages(opts.daemon_mode) != 0;
 
-	if (opts.mode == CR_CHECK)
+	case CR_CHECK:
 		return cr_check() != 0;
 
-	if (opts.mode == CR_PAGE_SERVER)
+	case CR_PAGE_SERVER:
 		return cr_page_server(opts.daemon_mode, false, -1) != 0;
 
-	if (opts.mode == CR_SERVICE)
+	case CR_SERVICE:
 		return cr_service(opts.daemon_mode);
 
-	if (opts.mode == CR_DEDUP)
+	case CR_DEDUP:
 		return cr_dedup() != 0;
 
-	if (opts.mode == CR_CPUINFO) {
-		if (!argv[optind + 1]) {
-			pr_err("cpuinfo requires an action: dump or check\n");
-			goto usage;
-		}
-		if (!strcmp(argv[optind + 1], "dump"))
-			return cpuinfo_dump();
-		else if (!strcmp(argv[optind + 1], "check"))
-			return cpuinfo_check();
-	}
+	case CR_CPUINFO_DUMP:
+		return cpuinfo_dump();
 
-	if (opts.mode == CR_EXEC_DEPRECATED) {
+	case CR_CPUINFO_CHECK:
+		return cpuinfo_check();
+
+	case CR_EXEC_DEPRECATED:
 		pr_err("The \"exec\" action is deprecated by the Compel library.\n");
 		return -1;
-	}
 
-	if (opts.mode == CR_SHOW_DEPRECATED) {
+	case CR_SHOW_DEPRECATED:
 		pr_err("The \"show\" action is deprecated by the CRIT utility.\n");
 		pr_err("To view an image use the \"crit decode -i $name --pretty\" command.\n");
 		return -1;
-	}
 
-	pr_err("unknown command: %s\n", argv[optind]);
+	case CR_UNSET:
+	default:
+		pr_err("unknown command: %s\n", cmd);
+	}
 usage:
 	pr_msg("\n"
 	       "Usage:\n"
@@ -429,7 +438,7 @@ usage:
 	       "  --network-lock METHOD network locking/unlocking method; argument\n"
 	       "                        can be 'nftables' or 'iptables' (default).\n"
 	       "  --unprivileged        accept limitations when running as non-root\n"
-	       "                        consult documentation for further details\n"
+	       "  --allow-uprobes       allow dump/restore with uprobes vma\n"
 	       "\n"
 	       "* External resources support:\n"
 	       "  --external RES        dump objects from this list as external resources:\n"
